// pulser - airborne galloping ghost decoder/driver
// ceptimus.  Started May 2018  Modifed August 2018 to auto-detect if elevator and aux channels are connected and work with/without them automatically
// Original was for ATmega328. This version for ATtiny85 February 2019.  Burn bootloader for internal 16Mhz no crystal.
// rudder input from receiver on Pin5, elevator on Pin6, auxilliary on Pin7 - these are actual package pin numbers, Pin5 is PB0, Pin6 is PB1, Pin7 is PB2
// AUX channel may be replaced by a potentiometer depending on configuration.
// all inputs have range 1ms = 0% to 2ms = 100%  Note: Taranis + Multimodule + Spektrum receiver only about 1120 - 1920 us range.
// rudder input 0% = 100% left, 100% = 100% right.  Typical usable range from 30% to 70%
// elevator input 0% = full down, fastest pulsing rate of 12 cycles per second.  100% = full up, slowest pulsing rate of 2 cycles per second.
// auxilliary input sets percentage of time motor H-bridge is powered (total of left and right phases) from 0% to 100%.

// leave #define in place for L9110(S) and other H-bridges that have active low inputs.  Comment out for MX1508, L298, TB6612FNG and other H-bridges with active high inputs 
// #define L9110

// default value to use for 'AUX' (drive power) input when input is not connected. 1800 gives about 90% power, which is good for magnetic actuators. 1400 might be better for some galloping ghost motor driven set ups
#define DEFAULT_AUX 1800

// default value to use for 'ELE' (pulse rate) input when input is not connected. 1300 gives a fairly fast pulse rate
#define DEFAULT_ELE 1300

#ifdef L9110
  #define OFF HIGH
  #define ON LOW
#else
  #define OFF LOW
  #define ON HIGH
#endif

#include <EEPROM.h>
#include "receiver.h"

// actual DIL package pin numbers for output drive: right is Pin2 (PB3) left is Pin3 (PB4)
#define RIGHT 3
#define LEFT 4

uint32_t remaining, eventTime;

uint8_t configuration; // 1 = pulsed with AUX control, 2 = pulsed with preset pot control, 3 = PWM with AUX control, 4 = PWM with preset pot control.

void setup(void) {
  // you can put your OSCCAL value for your particular board in the line below (and remove the // to enable it) if you know the correct value
  // if I've calibrated your board, I've written the value on the bottom.  It's two hex digits, so if your board has, say, B7 written on it
  // you put the two digits after the Ox on the line below - so it would say OSCCAL = 0xB7;
  // OSCCAL = 0x; 
  // if you don't know the OSCCAL value, don't guess - the value is then read from the ATtiny's manufacturer-set 'signature row'
  // this is different for every chip and is set by Atmel during testing - and will be pretty close.
  // but greater accuracy can still be obtained by calibration.

  uint8_t osccal = EEPROM.read(1); // has correct calibrated OSCCAL value for this board been stored in EEPROM since most recent reprogramming?
  if (osccal != 0xFF) { // if EEPROM doesn't contain the default value of an un-initialized EEPROM byte
    OSCCAL = osccal; // use the value for OSCCAL.  Idea is to get internal clock as near to 16 MHz as possible for most accurate reading of radio channel values.
  }
  
  digitalWrite(RIGHT, OFF); // set both H-bridge pins off for idle (coast) state 
  pinMode(RIGHT, OUTPUT);
  digitalWrite(LEFT, OFF);
  pinMode(LEFT, OUTPUT);
  
  // check if jumper(s) applied to select configuration
  // check if either pin zero and/or pin two linked to pin 1, also if pins zero and/or pin two linked to ground
  pinMode(0, INPUT_PULLUP);
  pinMode(1, OUTPUT);
  digitalWrite(1, LOW);
  pinMode(2, INPUT_PULLUP);
  bool connectedZeroOne = true;
  bool connectedTwoOne = true;
  bool connectedZeroGround = true;
  bool connectedTwoGround = true;
  for (uint8_t i=0; i<200; i++) {
    digitalWrite(1, LOW);
    delayMicroseconds(100);
    if (digitalRead(0) == HIGH) {
      connectedZeroGround = false;
      connectedZeroOne = false;
    }
    if (digitalRead(2) == HIGH) {
      connectedTwoGround = false;
      connectedTwoOne = false;
    }
    digitalWrite(1, HIGH);
    delayMicroseconds(100);
    if (digitalRead(0) == HIGH) {
      connectedZeroGround = false;
    } else {
      connectedZeroOne = false;
    }
    if (digitalRead(2) == HIGH) {
      connectedTwoGround = false;
    } else {
      connectedTwoOne = false;
    }
  }
  pinMode(1, INPUT_PULLUP);
  configuration = 0; // default
  if (connectedZeroOne) {
    configuration = 1;
    if (connectedTwoGround) {
      configuration = 3;
    }
  }
  if (connectedTwoOne) {
    configuration = 2;
    if (connectedZeroGround) {
      configuration = 4;
    }
  }
  if (connectedZeroOne && connectedTwoOne) { // special configuration 5 - for speed calibration
    configuration = 5;
  }
  if (configuration) { // write selected configuration to EEPROM
    if (configuration < 5) {
      EEPROM.update(0, configuration);
      EEPROM.update(1, OSCCAL);
    }
  } else { // read configuration from EEPROM
    configuration = EEPROM.read(0);
    if (configuration < 1 || configuration > 4) { // no valid configuration stored to EEPROM - default to configuration 1
      configuration = 1;
    }
  }

  if (configuration == 5) {
    pinMode(1, OUTPUT); // will be sending CLK / 1000 on ELE pin
    pinMode(0, OUTPUT); // will be serially transmitting current OSCCAL value on RUD pin
    eventTime = micros();
    return;
  }
  
  if (configuration == 1 || configuration == 3) {
    receiver::start(3); // start receiving three channels of data from radio receiver
  } else {
    receiver::start(2); // start receiving two channels of data from radio receiver
  }

  do {
    delay(10);
  } while (!receiver::active(0)); // wait for a valid receiver signal on the rudder channel

  if (configuration > 2) { // working in fast PWM mode - configure timer
    pinMode(RIGHT, INPUT); // disable both PWM outputs initially
    pinMode(LEFT, INPUT);
    TCCR1 = 0x94; // run at ~8 kHz PWM frequency.  0x93 gives ~16 kHz, 0x92 gives ~32 kHz, and 0x91 the fastest possible 62.5 kHz
    GTCCR = 0x50; // allow PWM B to control PB3 and PB4
    OCR1C = 0xFF; // count up to FF before resetting
    OCR1B = 0x80; // switch initially at half-way point
  }
  
  eventTime = micros();
}

// Special configuration 5 mode for calibrating oscillator speed - board outputs a 1000-Hz square wave on the 'ELE' pin
// and its current value of OSCCAL serially on the 'RUD' pin.  It listens for serial commands on its 'AUX' pin.
// A command of '+' increases the OSCCAL value (and therefore the speed), '-' reduces the speed, and 'W' asks the board
// to write its current OSCCAL value to EEPROM.  At subsequent power-on events the board reads and uses the stored value of
// OSCCAL from EEPROM.  Note that the default Arduino IDE 'Upload' using a USBasp or similar overwrites the EEPROM to all 'FF'
// The sketch recognizes the special FF value and doesn't use it for OSCCAL, but if you reprogram you need to recalibrate
// afterwards for greatest accuracy.  If you know the correct OSCCAL value for your chip, you can put it up near the top of
// this sketch and the board will then run using that value (and write that value to EEPROM the next time you change configuration.
// You can use my sketch running on a UNO, NANO, or similar and connect that board to your pulser - that causes the board to
// power up in this configuration mode - the current speed and OSCCAL value are displayed on the PC screen, and three buttons
// let you alter the speed and write it to EEPROM when it's correct.  The sketch is named pulserOSCCALcalibrate-UNO
void loopOSCCAL(void) {
  static uint8_t transmitByte, transmitMode = 0;
  static uint8_t receiveByte, receiveMode = 0;
  do { remaining = eventTime - micros(); } while (remaining < 0x80000000UL); // wait till micros() > eventTime, allowing for the wrap that occurs when micros() counts past 0xFFFFFFFF every 71 minutes 35 seconds (if flight should ever last that long!)
  digitalWrite(1, HIGH);
  // change written bits on positive edge
  if (transmitMode == 0) { // send start bit
    digitalWrite(0, HIGH);
    transmitByte = OSCCAL;
    transmitMode++;
  } else if (transmitMode <= 8) { // transmit bits of transmitByte, MSB first
    if (transmitByte & 0x80) {
      digitalWrite(0, HIGH);
    } else {
      digitalWrite(0, LOW);
    }
    transmitByte <<= 1;
    transmitMode++;
  } else if (transmitMode < 18) { // send 9 stop bits
    digitalWrite(0, LOW);
    transmitMode++;
  } else {
    transmitMode = 0;
  }
  eventTime += 500UL; // output 1000 Hz square wave on PB1 (assuming clock is running at the correct 16MHz)
  do { remaining = eventTime - micros(); } while (remaining < 0x80000000UL);
  digitalWrite(1, LOW);
  // bits change on positive edge and are read on negative edge
  uint8_t receiveState = 0;
  if (digitalRead(2) == HIGH) {
    receiveState = 0x01;
  }
  if (!receiveMode) {
    if (receiveState) { // start bit received
      receiveMode = 1;
      receiveByte = 0;
    }
  } else {
    if (receiveMode <= 8) { // receiving bits, MSB first
      receiveByte = (receiveByte << 1) | receiveState;
      receiveMode++;
    } else if (receiveMode == 9) { // expecting stop bit
      if (!receiveState) { // stop bit received
        switch (receiveByte) {
          case '+':
            OSCCAL++;
            break;
          case '-':
            OSCCAL--;
            break;
          case 'W':
            EEPROM.update(1, OSCCAL);
            break;
        }
      }
      receiveByte = receiveMode = 0; // back to idle mode
    }
  }
  eventTime += 500UL;  
}

void loop(void) {
  if (configuration == 5) {
    loopOSCCAL();
    return;
  }
  uint16_t power;
  // calculations of next set of event times all happen in the final 'coast' stage set up by the previous loop
  uint16_t cycle = receiver::active(1) ? receiver::channel(1) : DEFAULT_ELE; // elevator input: nominally 1000 - 2000 = 0 - 100%
  if (cycle < 1500) { // tweak input range to alter range 1000:1500:2000 to about 1100:1325:1800
    cycle = 1325 - ((1500 - cycle) >> 1);
  } else {
    cycle -= 175;
  }
  cycle = cycle > 1000 ? cycle > 2000 ? 1200 : cycle - 800 : 200; // shift and clip to 200 - 1200 range to give 6:1 ratio between fast and slow
  uint32_t cycleMicros = 417UL * cycle; // cycleMicros is length of one complete pulse cycle in microseconds.  Range 83400 to 500400 - about 12 Hz to 2 Hz

  if (configuration & 0x0001) { // reading AUX channel for power control
    power = !receiver::active(0) ? 1125 : receiver::active(2) ? receiver::channel(2) : DEFAULT_AUX; // auxilliary input: nominally 1125 - 1875 = 0 - 100% power but default to zero power if rudder signal not active
  } else { // reading pot for power control
    power = !receiver::active(0) ? 1125 : analogRead(1) + 1125;
  }
  power = power > 1125 ? power > 1875 ? 750 : power - 1125 : 0; // shift and clip to range 0 - 750

  uint16_t right = receiver::channel(0);  // amount of 'right': 1000 = 0% right to 2000 = 100% right
  right = right > 1000 ? right > 2000 ? 1000 : right - 1000 : 0; // shift and clip to range 0 - 1000  ************* -980 gives better centering with test pulser board (one with ICSP header) *************
  right = right >> 1; // divide by two to give range 0 - 500 for 0% to 100% right
  uint16_t left = 500 - right; // left + right always totals 500


  if (configuration <= 2) { // running in pulse mode
    power /= 3; // divide by three to give 0 - 250
    uint32_t poweredMicros = (cycleMicros >> 8) * power; // divide total cycle time by 256 and multiply by power to give a (total of left and right) powered duration of 0 to 250/256 of the total cycle time.
    uint32_t powerDelta = poweredMicros >> 9; // powerDelta is 512 times shorter than poweredMicros
    uint32_t rightMicros = powerDelta * right; // 'right' output duration
    uint32_t leftMicros = powerDelta * left; // 'left' output duration
    
    uint32_t coastInterval = (cycleMicros - poweredMicros) >> 1; // two equal-sized coast intervals pad out the periods from (1) switching off 'right' to switching on 'left' and (2) switching off 'left' till switching on 'right' in the next cycle
  
    if (coastInterval < 250) { // allow minimum coastInterval of 250us - the loop overhead plus above code (worst case analaog input) takes up to 212us
      uint32_t subtract = 250 - coastInterval;
      if (rightMicros > subtract) {
        rightMicros -= subtract; 
      } else {
        rightMicros = 0;
      }
      if (leftMicros > subtract) {
        leftMicros -= subtract;
      } else {
        leftMicros = 0;
      }
      coastInterval = 250;
    }
    do { remaining = eventTime - micros(); } while (remaining < 0x80000000UL); // wait till micros() > eventTime, allowing for the wrap that occurs when micros() counts past 0xFFFFFFFF every 71 minutes 35 seconds (if flight should ever last that long!)
    // output the LEFT and RIGHT phases
    if (receiver::active(0)) {
      digitalWrite(RIGHT, ON);
      eventTime += rightMicros;
      do { remaining = eventTime - micros(); } while (remaining < 0x80000000UL);
      digitalWrite(RIGHT, OFF);
      eventTime += coastInterval;
      do { remaining = eventTime - micros(); } while (remaining < 0x80000000UL);
      digitalWrite(LEFT, ON);
      eventTime += leftMicros;
      do { remaining = eventTime - micros(); } while (remaining < 0x80000000UL);
      digitalWrite(LEFT, OFF);
      eventTime += coastInterval;
    } else {
      digitalWrite(RIGHT, OFF);
      digitalWrite(LEFT, OFF);
      eventTime += rightMicros;
      eventTime += coastInterval;
      eventTime += leftMicros;
      eventTime += coastInterval;
    }
  } else { // running in fast PWM mode
    uint8_t ocr1bVal;
    uint32_t switchInterval = 250;
    cycleMicros -= 500; // allow for two switching intervals
    cycleMicros /= 500; // prescale before multiplying by (left + right), which always total 500
    uint32_t rightMicros = cycleMicros * right;
    uint32_t leftMicros = cycleMicros * left;
    uint32_t powerX256 = ((uint32_t)(power)) << 8;
   
    power = (uint16_t)(powerX256 / 750); // power is now in range 0-256
    if (power > 255) {
      power = 255;
    }
    
    do { remaining = eventTime - micros(); } while (remaining < 0x80000000UL); // wait till micros() > eventTime, allowing for the wrap that occurs when micros() counts past 0xFFFFFFFF every 71 minutes 35 seconds (if flight should ever last that long!)
    pinMode(RIGHT, OUTPUT); // enable PWM on right output
    eventTime += rightMicros;
    do { remaining = eventTime - micros(); } while (remaining < 0x80000000UL);
    pinMode(RIGHT, INPUT); // disable PWM on right output
    digitalWrite(RIGHT, LOW);
    ocr1bVal = (uint8_t)(power);
    OCR1B = ocr1bVal;
    eventTime += switchInterval;
    do { remaining = eventTime - micros(); } while (remaining < 0x80000000UL);
    pinMode(LEFT, OUTPUT); // enable PWM on left output
    eventTime += leftMicros;
    do { remaining = eventTime - micros(); } while (remaining < 0x80000000UL);
    pinMode(LEFT, INPUT); // disable PWM on left output
    digitalWrite(LEFT, LOW);
    ocr1bVal = (uint8_t)(255  - power );
    OCR1B = ocr1bVal;
    eventTime += switchInterval;
  }
}
