#include "receiver.h"

uint8_t receiver::numChannels; // set to 2 or 3 by start method
uint8_t receiver::prev; // remembers state of input bits from previous interrupt
uint32_t receiver::risingEdge[3]; // time of last rising edge for each channel
volatile uint32_t receiver::uSec[3]; // the latest measured pulse width in uS for each channel

void receiver::start(uint8_t nc) { // nc = numChannels should be 2 or 3
  numChannels = nc;
  for (int pin = 0; pin < numChannels; pin++) { // enable PB0, PB1, (PB2) (physical pins 5,6,(7)) as our 2 or 3 input bits
    pinMode(pin, INPUT_PULLUP);
  }
  for (int n = 0; n < 3; n++) {
    risingEdge[n] = 0x80000000UL; // default rising edge times so that active() returns false before a signal is received
  }

  if (numChannels == 3) {
    PCMSK |= 0x07; // set the mask to allow 3 pins to generate interrupts
  } else {
    PCMSK |= 0x03; // set the mask to allow 2 pins to generate interrupts
    digitalWrite(2, LOW);
    pinMode(2, INPUT); // configure PB2 to work as analog input    
  }
  GIMSK |= 0x20;  // enable pinchange interupts
}

uint16_t receiver::channel(uint8_t n) { // return uS for channel[n]. n is channel 0 to 2.
   uint16_t uS;
   cli();
   uS = (uint16_t)uSec[n];
   sei();
   if (uS < 1000)
    uS = 1000;
   else if (uS > 2000)
    uS = 2000;
   return uS;
}

bool receiver::active(uint8_t n) { // returns true if channel[n] has received +ve edge in last half-second
  uint32_t uS = micros();
  cli();
  uS -= risingEdge[n];
  sei();
  return uS < 500000UL;
}

ISR(PCINT0_vect) { // one or more of pins 5, 6, (7) have changed state
  uint8_t curr;
  uint32_t now = micros();
  if (receiver::numChannels == 3) {
    curr = PINB & 0x07; // current state of the 3 input bits
  } else {
    curr = PINB & 0x03; // current state of the 2 input bits    
  }
  uint8_t changed = curr ^ receiver::prev;
  uint8_t mask = 0x01;
  for (uint8_t channel = 0; channel < receiver::numChannels;  channel++) {
    if (changed & mask) { // this pin has changed state
      if (curr & mask) { // +ve edge so remember time
        receiver::risingEdge[channel] = now;
      }
      else { // -ve edge so store pulse width
        receiver::uSec[channel] = now - receiver::risingEdge[channel];
      }
    }
    mask <<= 1;
  }
  receiver::prev = curr;
}
